/*------------------------------------------------------------------------
Module:        mime_alter.c
Author:        Paul L Daniels / Steven Coutts
Project:       ripMIME
State:         development
Creation Date: 01/05/2001
Description:   MIME-alter is an additional module in the mime.[ch] object, which will contain various functions which will -alter- mime-packs.
------------------------------------------------------------------------*/

// 0908 10H57 - Added ability for disclaimer to be actual text, as well as
// 	a filename.  I just detect for a \n or \r in the string, if detected
//		disclaimer is considered as "text"
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <syslog.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "ffget.h"
#include "mime_alter.h"

/* Our content types */
#define _unknown -1
#define _none 0
#define _text_plain 1
#define _text_html 2
#define _text_print 3
#define _text_multipart 4
#define _text_unknown 5

/* Our encoding scemes */
#define _base64 1
#define _print 2
#define _7bit 3

#define _BUFFER_MAX 1024

/* Globals */
char *buffer[_BUFFER_MAX]; /* allocate a static array of 1024 pointers to char */


//06/03/2002 - replaced AM_setup_table() with this compile-time array. Saves a few startup cycles.

unsigned char encode64[64]={
	65,   66,   67,   68,   69,   70,   71,   72,   73,   74,   75,   76,   77,   78,   79,   80,\
  81,   82,   83,   84,   85,   86,   87,   88,   89,   90,   97,   98,   99,  100,  101,  102,\
 103,  104,  105,  106,  107,  108,  109,  110,  111,  112,  113,  114,  115,  116,  117,  118,\
 119,  120,  121,  122,   48,   49,   50,   51,   52,   53,   54,   55,   56,   57,   43,   47 \
};


static int __debug=0;				// Low level debugging

static int __verbose=0;				/* do we talk as we walk */
static int __paranoid=1;				/* set paranoid to yes! */
static int __HTMLtoo=1;				/* Add footer to the HTML email too */
static int __nullify_all=0;			/* Remove ALL filename'd attachments */
static int __alter_signed=0;		/* Do we alter signed emails ? */

int _AM_syslogging=1;		/* do we want to log to syslog */
int _AM_stderrlogging=0;	/* do we want to log to stderr */

#define _HEADERBUFFER_MAX 100
char *headerbuffer[_HEADERBUFFER_MAX];	// 100 lines for the header buffers

int headerbuffermax=0;
int _debugAM = 0;




/*------------------------------------------------------------------------
Procedure:     AM_set_debug ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_debug( int level )
{
	_debugAM = level;
	__debug = level;
	AM_set_verbose( level );

	return level;

}


/*------------------------------------------------------------------------
Procedure:     AM_hbuffer_reset ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_hbuffer_reset( void )
{
	int i;

	for (i = 0; i < headerbuffermax; i++)
	{
		if (headerbuffer[i]) free(headerbuffer[i]);
	}
	headerbuffermax=0;
	return 0;
}


/*------------------------------------------------------------------------
Procedure:     AM_hbuffer_add ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
char *AM_hbuffer_add( char *headerline, FILE *f )
{
	int i;

	if (headerbuffermax == _HEADERBUFFER_MAX)
	{
		if (headerbuffer[0])
		{
			fprintf(f,"%s",headerbuffer[0]);
			free(headerbuffer[0]);
		}
		for (i = 0; i < (_HEADERBUFFER_MAX-1); i++)
		{
			headerbuffer[i] = headerbuffer[i+1];
		}
		headerbuffermax = _HEADERBUFFER_MAX -1;
	}

	headerbuffer[headerbuffermax] = malloc(sizeof(char) *(1025));

	if (!headerbuffer[headerbuffermax])
	{
		syslog(1,"alterMIME: AM_hbuffer_add: Error: cannot allocate %d bytes for new header",sizeof(headerline)+1);
		return NULL;
	}

	strncpy(headerbuffer[headerbuffermax], headerline, 1024);
	headerbuffermax++;

	return headerbuffer[headerbuffermax-1];
}


/*------------------------------------------------------------------------
Procedure:     AM_hbuffer_getmax ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_hbuffer_getmax( void )
{
	return headerbuffermax;
}




/*------------------------------------------------------------------------
Procedure:     AM_hbuffer_getline ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
char *AM_hbuffer_getline( int index )
{
	if ((index >=0)&&(index < headerbuffermax))
		return headerbuffer[index];
	else return NULL;
}



/*------------------------------------------------------------------------
Procedure:     AM_set_paranoid ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_paranoid( int level )
{
	__paranoid = level;
	return __paranoid;
}





/*------------------------------------------------------------------------
Procedure:     AM_set_verbose ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_verbose( int level )
{
	__verbose = level;
	return __verbose;
}




/*------------------------------------------------------------------------
Procedure:     AM_set_HTMLtoo ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_HTMLtoo( int level )
{
	__HTMLtoo = level;
	return __HTMLtoo;
}




/*------------------------------------------------------------------------
Procedure:     AM_set_nullifyall ID:1
Purpose:
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_nullifyall( int level )
{
	__nullify_all = level;
	return __nullify_all;
}



/*------------------------------------------------------------------------
Procedure:     AM_set_altersigned ID:1
Purpose:       Turns on or off the option to alter signed email messages.
Default is to be -off-.
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_set_altersigned( int level )
{
	__alter_signed = level;
	return __alter_signed;
}




/*------------------------------------------------------------------------
Procedure:     strlower ID:1
Purpose:       Converts a given string into all lowercase
Input:         convertme : String to be converted
Output:        none
Errors:
------------------------------------------------------------------------*/
int AM_strlower( char *convertme )
{

	char *c = convertme;

	while ( *c != '\0') {
		*c = tolower(*c);
		c++;
	}

	return 0;
}




/*------------------------------------------------------------------------
Procedure:     AM_create_encode_table ID:1
Purpose:       Builds the value:ASCII relationship table for BASE64 -encoding-
process
Input:         none
Output:
Errors:
------------------------------------------------------------------------*/
int AM_create_encode_table( void )
{

	fprintf(stderr,"AM_create_encode_table:%d: WARNING - Depreated function. Please do not use.\n",__LINE__);
	return 0;

	/*
	int i,j;

	for ( i = 'A', j=0; i <= 'Z'; i++, j++)
	{
	encode64[j]=i;
	}

	for ( i = 'a'; i <= 'z'; i++, j++)
	{
	encode64[j]=i;
	}

	for ( i = '0'; i <= '9'; i++, j++)
	{
	encode64[j]=i;
	}

	encode64[62] = '+';
	encode64[63] = '/';
	return 0;
	*/
}










/*------------------------------------------------------------------------
Procedure:     AM_base64_encodef ID:1
Purpose:       encode to BASE64 an input stream to an output stream
Input:         FILE *fin: Input stream
FILE *fout: Output stream
Output:
Errors:
------------------------------------------------------------------------*/
int AM_base64_encodef( FILE *fin, FILE *fout )
{
	unsigned char inbuf[3];
	unsigned char outbuf[4];
	int cc;
	int byte_count;

	if (!fin)
	{
		syslog(1,"AM_base64_encodef: Error: input file stream not open, please use AM_base64_encode(infile,outfile)");
		return -1;
	}

	if (!fout)
	{
		syslog(1,"AM_base64_encodef: Error: output file stream not open, please use AM_base64_encode(infile,outfile)");
		return -1;
	}

	AM_create_encode_table(); /* setup the binary->base64 encoding table */

	cc = 0;	/* Character Count = 0 */

	while ((byte_count = fread(inbuf, 1, 3, fin)))	/* while we get more than 0 bytes */
	{
		/* Split 3 bytes into 4 bytes by taking 6bit blocks out of the 24 bit (3x8bit)
		* array */
		outbuf[0] = encode64[((inbuf[0] & 0xFC) >> 2)];
		outbuf[1] = encode64[((inbuf[0] & 0x03) << 4) | ((inbuf[1] & 0xF0) >> 4)];
		outbuf[2] = encode64[((inbuf[1] & 0x0F) << 2) | ((inbuf[2] & 0xC0) >> 6)];
		outbuf[3] = encode64[(inbuf[2] & 0x3F)];

		/* if we didn't get a full 3 bytes from the file read, then we need to then
		* pad the output with '=' (base64 padding char) */
		if ( byte_count < 3 )
		{

			if ( byte_count == 2 ) outbuf[3] = '=';
			if ( byte_count == 1 ) outbuf[3] = outbuf[2] = '=';

			fwrite(outbuf, 1, 4, fout); 	/* write out the buffer */
			fwrite( "\n", 1, 1, fout ); 	/* as this is the last line, put a trailing \n */
			break; 								/* exit out of the while loop */
		}

		fwrite(outbuf, 1, 4, fout);		/* normal operation buffer-write */

		cc += 4;									/* increment the output char count by 4 */

		if (cc > 76)							/* if we have more than 76 chars, then put in a \n */
		{
			fwrite( "\n", 1, 1, fout );	/* write the \n out */
			cc = 0;								/* reset the output char count */
		} /* if bytecount == 3 */
	} /* while more data to read in */

	return 0;
}






/*------------------------------------------------------------------------
Procedure:     AM_base64_encode ID:1
Purpose:       Encodes a given input stream into BASE64 and outputs to a given
output stream
Input:         char *enc_fname: Input filename of the stream to encode
char *out_fname: Output filename of the stream to send to
Output:
Errors:
------------------------------------------------------------------------*/
int AM_base64_encode( char *enc_fname, char *out_fname )
{
	FILE *fin, *fout;

	fin = fopen( enc_fname, "rb" );
	if (!fin)
	{
		syslog(1,"AM_base64_encode: Cannot open \"%s\" for reading.(%s)",enc_fname,strerror(errno));
		return -1;
	}

	fout = fopen ( out_fname, "wb" );
	if (!fout)
	{
		syslog(1,"AM_base64_encode: Cannot open \"%s\" for writing.(%s)",out_fname,strerror(errno));
		fclose(fin);
		return -1;
	}

	AM_base64_encodef( fin, fout );	/* encode and output */

	fclose(fin);	/* close the input file */
	fclose(fout);	/* close the output file */

	return 0;
}




/*------------------------------------------------------------------------
Procedure:     AM_quick_clean_filename ID:1
Purpose:       scans all the chars of a filename, converts any non-safe chars to
a '_'
Input:         char *filename: The filename we're going to convert/scan
Output:
Errors:
------------------------------------------------------------------------*/
void AM_quick_clean_filename( char *fname )
{
	char tmp[1024];
	char *p = strrchr(fname,'/');

	/* scan out any directory separators */
	if (p)
	{
		p++;
		sprintf(tmp,"%s",p);
		strcpy(fname,p);
	}
	else if ( (p = strrchr(fname,'\\')))
	{
		p++;
		sprintf(tmp,"%s",p);
		strcpy(fname,p);
	}


	while (*fname) {
		if ( !isalnum(*fname)||(*fname == ' ') ) *fname='_';
		fname++;
	}
}






/*------------------------------------------------------------------------
Procedure:     AM_attachment_replace ID:1
Purpose:       Replace the attachment in a mailbody
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int AM_attachment_replace( char *mpack_name, char *old_att_name, char *new_att_name, char *att_path )
{
	FILE *in, *tmp, *newa;
	char *ofn, *nfn;
	char line[1024];
	char lline[1024];
	char lname[1024];
	char ts[1024];
	char tname[1024];
	char *p, *q;
	int base64hit = 0;
	int hit = 0;
	int quote_on = 0;
	int namesmatch = 0;
	int i = 0;
	char *index;
	char *nameindex = NULL;
	char *fnameindex = NULL;

	in = fopen( mpack_name, "r" ); /* open up our original mailpack */

	if (!in)	/* handle file opening error */
	{
		syslog(1,"AM_attachment_replace: Cannot open mailpack %s for reading, \"%s\""\
		,mpack_name\
		,strerror(errno));
		return -1;
	}


	// First we must take the old file name, and translate it into
	// something we can use to compare against after it has been
	// brutalised and reduced (for safety)
	//
	ofn = strrchr(old_att_name,'/');  	/* get the attachment name out of the full path */
	if (!ofn) ofn = old_att_name;			/* if there was no '/', then reset the ofn ptr */
	else ofn++; 								/* skip the last '/' */
	strcpy(lname, ofn); 						/* create a lowercase version */
	AM_strlower(lname);
	AM_quick_clean_filename(lname); 		/* clean up the filename */

	// Get the new attachment name out of a full path
	// This is simply so that we can insert this new name
	// into the headers without the pointless path prefix.
	//
	nfn = strrchr(new_att_name,'/');		/* extract new att name from path */
	if (!nfn) nfn = new_att_name;			/* if there was no '/' then reset the nfn ptr */
	else nfn++;


	if (_debugAM) syslog(1,"AM_replace(): looking to replace %s with %s, having %s and %s", ofn, nfn, mpack_name, att_path );

	// Create the temporary mailpack name
	// This will become the new mail pack
	// assuming that everything goes okay in the end.
	//
	sprintf(tname,"%s.tmp", mpack_name);/* create the temporary mailpack name */
	tmp = fopen( tname, "w" );				/* try open the temp file */
	if (!tmp) /* handle file opening error */
	{
		syslog(1,"AM_attachment_replace: Cannot open temporary file %s for writing, \"%s\""\
		,tname\
		,strerror(errno));
		return -1;
	}


	// Into the mailpack, first we must read in the headers, this must
	// be done in memory, so that should we encounter any of the required
	// header details, we dont have to try and back-track through a file
	// This gets rather awkward, as we dont know (often) if the file is
	// replacable before we get to its first occurance of the filename
	// header. This means we often have to backtrack ( a real PAIN ).
	//
	while (!feof(in))
	{
		hit = 0;
		base64hit = 0;
		quote_on = 0;
		namesmatch = 0;
		nameindex = NULL;
		fnameindex = NULL;
		AM_hbuffer_reset();// prepare the header buffers


		// Read in our file lines, searching for our attachment
		// which we are supposed to be replacing.
		//
		while (fgets(line, 1024, in))
		{

			// Store this line into the header-buffer for later
			// use.
			//
			index = AM_hbuffer_add(line, tmp);

			strcpy(lline,line); /* copy and convert to lowercase the line we've read */
			AM_strlower(lline);

			// If we detect "base64" on this line, then set the flag
			//
			if (strstr(lline,"base64"))
			{
				if (_debugAM) syslog(1,"AM_replace(): BASE64 found - flag set.");
				base64hit=1;
			}

			// If we detect "filename=" on this line, then we need to
			// remember it, for later processing (if required)
			//
			if ((p=strstr(lline,"name="))) /* if the line contains name=, then check it */
			{
				if (!fnameindex) fnameindex = index;
				else nameindex = index;

				// Okay, so we've found the "name" line, now, lets
				// process the data on this line, and see if we have a
				// matching filename.
				//
				quote_on = 0;					/* no quotes found yet */
				p = p +strlen("name=");		/* move to just after the ='s */
				if (*p == '\"')				/* if there's a quote... */
				{
					quote_on++;					/* set quote to ON */
					p++;							/* move to next char */
				}

				q = strtok(p,";\"\n\r");	/* extract the filename */
				AM_quick_clean_filename(q);/* clean the filename */

				// Compare both the length and contents of the
				// filenames.
				//
				if (\
					(strlen(q) == strlen(lname))\
					&& (strncmp(q,lname, strlen(lname))==0)\
					)
				{
					if (_debugAM) syslog(1,"AM_replace(): Hit on filenames  %s and %s",q,lname);
					namesmatch = 1;
				}
				else
				    {
					namesmatch = 0;
					base64hit = 0;
					fnameindex = 0;
					nameindex = 0;
				}
			}


			// Now, if we've found this is a base64 attachment and we
				// have also matched the names (matching done here)
				//
			if ( (base64hit == 1) && (namesmatch == 1) )
			{
				if (_debugAM) syslog(1,"Base64 and NameMatch located, now fixing headers");
				// Okay, so we've gotten the match, and we need to replace
					// this filename details with those of the new file...
					// This is were we use the nameindex and fnameindex, for
					// from those two points, we alter the filenames
					//

				// First, lets do the "name" only one
					// strcpy the details so we can play with it safely
					//
				if (nameindex)
				{
					strcpy(line, nameindex);
					AM_strlower(line);

					// 2. find where the "name" bit starts, and \0 it there
					//
					p = strstr(line,"name=") +strlen("name=");

					// 3. If the filename was quoted, remember this, and move one
					// 	char off to the right.
					if (*p == '"')
					{
						p++;
						quote_on=1;
					}
					else quote_on=0;
					*p = '\0';

					// 4. The actual filename is now ahead of us, until the next
					//		quote or \n\r.  We in effect, can ignore this now, and
					//		construct our new line.
					//
					sprintf(lline,"%s%s",line,nfn);
					if (quote_on) strcat(lline,"\"");
					strcat(lline,"\n");

					// 5. Copy this created string back, this requies a little
					//
					strncpy(nameindex,lline,1023);
				}

				//
					// The end of the filename swapout procedure.
					//
					// Now, do the same for the "filename" line....
					//
					if (fnameindex)
				{
					strcpy(line, fnameindex);
					AM_strlower(line);
					p = strstr(line,"name=") +strlen("name=");
					if (*p == '"')
					{
						p++;
						quote_on=1;
					}
					else quote_on=0;
					*p = '\0';
					sprintf(lline,"%s%s",line,nfn);
					if (quote_on) strcat(lline,"\"");
					strcat(lline,"\n");
					strncpy(fnameindex,lline,1023);
				}
				// We are now happy with the way the headers are in the
					// memory, we can dump them to the output file now.
					//
					hit++;						/* flag that we had success */
			}

			if (hit && base64hit) break;


		} // WHILE fgets(line)

		// Write out the headers which we have been so faithfully
		// keeping in our holds
		//
		for (i = 0;
		i < AM_hbuffer_getmax();
		i++)
		{
			fprintf(tmp,"%s",AM_hbuffer_getline(i));
		}

		// Reset the buffers
		//
		AM_hbuffer_reset();


		/* If we had success with finding the right filename etc... */
		if ((hit)&&(base64hit))
		{
			hit = base64hit = 0;

			/* Step 2. Locate the start of the attachment */
			/* Step 2. Locate the start of the attachment */
			/* Step 2. Locate the start of the attachment */
			while (fgets(line, 1024, in))
			{

				strcpy(lline,line);			/* copy and convert the next line */
				AM_strlower(lline);

				if ((p=strstr(lline,"name=")))	/* if the line has name= in it... */
				{

					p = p +strlen("name="); /* get the position where the name starts */
					if (*p == '\"')			/* if quoted... */
					{
						quote_on=1;				/* set quote flag */
						p++;						/* step one char to the right */
					}
					else quote_on=0;

					q = strtok(p,";\"\n\r");	/* extract the filename */
					AM_quick_clean_filename(q);/* clean the filename */

					if ( (strlen(q) == strlen(lname))\
				&& (strncmp(q,lname, strlen(lname))==0)
						    )
					{
						p = q;
						q = line +(p -lline);
						*q = '\0';
						sprintf(ts, "%s%s", line, nfn);
						if (quote_on) strcat(ts,"\"");
						strcat(ts,"\n");
						strcpy(line,ts);
					}
				}


				fprintf(tmp,"%s",line);

				if (strlen(line) < 3) break;	/* if we've gotten to a blank line, exit */
			}


			/* Step 3. Insert attachment */
			/* Step 3. Insert attachment */
			/* Step 3. Insert attachment */

			newa = fopen( att_path, "r" );		/* Open the new attachment */

			if (!newa)										/* if we couldn't open it */
			{
				syslog(1,"AM_attachment_replace: Cannot open file \"%s\" required to be inserted.(%s)"\
			,new_att_name, strerror(errno));

				/* as we couldn't open the new attachment, then we better keep the
				* existing file so that we dont bugger up the mailpack, better to have
				* the old attachment than non at all! */
				while (fgets(line, 1024, in))
				{
					fprintf(tmp,"%s",line);
					if (strpbrk(line,"-")) break;
				} // While getting input file

			} // if New Attachment couldn't be opened

			else
			    {
				if (_debugAM) syslog(1,"AM_replace(): Starting to encode file...");
				AM_base64_encodef( newa, tmp );	/* Encode and output the new attachment */
				fclose(newa);								/* close the new attachment */
				if (_debugAM) syslog(1,"AM_replace(): Encoding and insertion complete.");

				/* absorb the old attachment */
				while (fgets(line, 1024, in))
				{
					/* we're assuming that in base64, only
					* boundary lines have a '-' in them
					* This is possibly a bad thing (IMHO)
					* but short of tracking boundaries, I
					* have no other way */
					if (strpbrk(line,"-"))
					{
						fprintf(tmp,"%s",line);
						break;
					} // If strpbrk

				} // While fgets

				if (_debugAM) syslog(1,"AM_replace(): Attachment absorption complete.");

			} // ELSE

		} // While (major) fgets

	} // While !feof()

	clearerr(in);										/* clear out any errno's */
	fclose(tmp);										/* close our new mailpack */
	fclose(in);											/* close our old mailpack */
	rename( tname, mpack_name );					/* rename! */

	return 0;
}






/*------------------------------------------------------------------------
Procedure:     AM_add_section1 ID:1
Purpose:       Reads the main headers off an email and places the values it finds for
content-type, encoding and other such into the dd record
Input:         struct AM_disclaimer_details *dd : pointer to the record where all the details will go
FFGET_FILE *f : Open stream which we're reading the headers from
FILE *newf : Open stream which we're dumping the headers we read back into.
Output:
Errors:
------------------------------------------------------------------------*/
int AM_add_section1( struct AM_disclaimer_details *dd, FFGET_FILE *f, FILE *newf )
{
	char line[_BUFFER_MAX +1];
	char lline[_BUFFER_MAX +1];
	char *pos, *newpos, *tmppos;

	dd->isfile = 0;
	dd->ishtml = 0;

	if (FFGET_feof(f)) return -1;

	while ( FFGET_fgets(line, _BUFFER_MAX +1, f) )
	{
		fprintf(newf,"%s",line);
		if (__debug) fprintf(stderr,"%s",line);
		if (f->trueblank)
		{
			// If the line read in was a /TRUE-BLANK/, then we have
			// reached the end of our headers, so, it's time to
			// move on to the next section.

			break;
		}

		strcpy(lline,line);
		AM_strlower(lline);

		// Look for various header components in the lines we just read in
		//

		// Test for a filename

		if ( (strstr(lline,"name="))||(strstr(lline,"name =")) )
		{
			dd->isfile = 1;
		}

		// Test for content-type
		//
		// We need to determine the content-type so that we can test to see if
		// we've got a plain-text or a HTML body involved
		//

		if ((pos=strstr(lline, "content-type:")))
		{

			if (strstr(lline,"/signed"))
			{
				if (__alter_signed == 0)
				{
					if (__verbose) fprintf(stderr,"Message is signed, not altering.\n");
					return AM_RETURN_SIGNED_EMAIL;
				}
			}

			pos += strlen("content-type:");	// move our pointer along

			tmppos = pos;
			if ((pos=strstr(tmppos,"text/")))	// do we have a text type
			{
				// this is a text section

				dd->content_type = _text_plain;

				// We honestly dont "care" about the complexities of the text included
				// other than if it's  HTML

				pos += strlen("text/");	// move pointer along

				if (strncmp(pos,"html",4)==0) dd->content_type = _text_html;

				/*
				if (strncmp(pos, "plain",5)==0) dd->content_type = _text_plain;
				else if (strncmp(pos, "html",4)==0) {
				dd->content_type = _text_html;
				dd->ishtml = 1;
				}
				else if (strncmp(pos, "print",5)==0) dd->content_type = _text_print;
				else dd->content_type = _text_unknown;
				*/
			}
			else if ((pos=strstr(tmppos,"multipart/")))
			{
				dd->content_type = _text_multipart;
			}
			else
			    {
				dd->content_type = _unknown;
			}

			if (__verbose) fprintf(stderr,"Content line found...\n");

		} // Content-type line



		// Test for Content-Transfer-Encoding:
		//
		// CTE indicates what sort of format the data is encoded in in the body
		// As we're only interested in dealing with /text/ based content, it's
		// important that we find out what encoding we have.
		//
		// There is little point trying to add a disclaimer to something which
		// is BASE64 encoded!
		//

		if ((pos=strstr(lline, "content-transfer-encoding:")))
		{
			pos += strlen("content-transfer-encoding:");
			if (strstr(pos,"base64")) dd->content_encoding = _base64;
			else if (strstr(pos,"quote")) dd->content_encoding = _print;
			else if (strstr(pos,"7bit")) dd->content_encoding = _7bit;
			if (__verbose) fprintf(stderr,"Content encoding found...\n");
		} // If line was a CTE




		// Test for Boundary specifier
		//
		// A very important line to detect for.  Without the boundary specifier, we
		// cannot correctly parse through a MIME encoded email.
		//

		if (dd->boundary_found == 0)
		{
			pos = strstr(lline, "boundary=");
			if (pos)
			{
				newpos = strstr(lline,"subject");
				if ((!newpos)||(newpos > pos))
				{
					if (__verbose) fprintf(stderr,"Possible boundary...");
					newpos = (pos -lline) +line +strlen("boundary=");	// move the pos pointer to just after the =

					pos = newpos;

					if ('\0' != *(pos +1))
					{
						if (*pos == '\"') pos++;		// If the boundary is quoted, then move along one char

						newpos = strtok(pos,"\n\r\"");	// Tokenise out the boundary string

						if (newpos)
						{
							strcpy(dd->boundary, newpos);		// copy it to our static string array

							dd->boundary_found = 1;				// Set the boundary_found flag

							if (__verbose) fprintf(stderr,"boundary located...");
						}
					}

				} // If boundary is -genuine-

			} // If boundary located

		} // If we don't already have a boundary


	} // Section 1 completed (Read headers)

	if (__verbose) fprintf(stderr,"%s:%d:Exiting headers read...\n\t-- isfile=%d ishtml=%d boundaryfound=%d\n\n",__FILE__,__LINE__, dd->isfile, dd->ishtml, dd->boundary_found );
	return 0;
}



/*------------------------------------------------------------------------
Procedure:     AM_disclaimer_load_text ID:1
Purpose:       Loads the text from the file given in fname, and mallocs the required amount
of memory to place the entire contents of the file into textptr.
Input:         char *fname : The file name from which to read the dislcaimer text
char **textptr : a derefernced pointer to the buffer variable which will hold the text.
Output:
Errors:
------------------------------------------------------------------------*/
int AM_disclaimer_load_text( char *fname, char **textptr )
{
	FILE *f;
	struct stat st;

	if (0 == stat(fname, &st))
	{
		*textptr = malloc(st.st_size +1);  // We have to add 1 so that we delimit the data with a \0

		if (textptr)
		{

			memset(*textptr, '\0', (st.st_size +1));

			f = fopen(fname,"r");
			if (!f)
			{
				fprintf(stderr,"%s:%d:ERROR - Cannot open %s for reading (%s)",__FILE__,__LINE__,fname,strerror(errno));
				textptr = NULL;
				return 0;
			}

			fread(*textptr, 1, st.st_size, f);
			if (__verbose) fprintf(stderr,"Disclaimer Loaded:\n%s",*textptr);
			fclose(f);
		}
	}
	else
	    {
		fprintf(stderr,"%s:%d:ERROR - Cannot stat '%s' (%s)",__FILE__,__LINE__,fname,strerror(errno));
	}

	return 0;
}




/*------------------------------------------------------------------------
Procedure:     AM_add_disclaimer_insert_html ID:1
Purpose:       Inserts a disclaimer with <PRE> tagging into a HTML block of text just prior to the
end of the HTML. Assumes that it has been placed at the start of the HTML block
Input:         AM_disclaimer_details *dd : Disclaimer tracking information
FFGET_FILE *f : Input file
FILE *newf : Output file
Output:
Errors:
------------------------------------------------------------------------*/
int AM_add_disclaimer_insert_html( 	struct AM_disclaimer_details *dd, FFGET_FILE *f, FILE *newf )
{
	char line[_BUFFER_MAX +1];
	char lline[_BUFFER_MAX +1];
	char *prebody, *tmpbody;

	while (FFGET_fgets(line, _BUFFER_MAX, f))
	{
		strcpy(lline,line);
		AM_strlower(lline);

		// Look for either a BODY or HTML closing tag.

		prebody = strstr(lline,"</body>");
		if (!prebody) prebody = strstr(lline,"</html>");

		// If we found one of the tags, then insert our disclaimer aruond about
							// here.

		if (prebody != NULL)
		{
			if (__verbose) fprintf(stdout,"Inserting html-body disclaimer...");

			dd->html_inserted = 1;

			// prepare to print out the original up to the </body> or </html> part

			tmpbody = line +(prebody -lline);
			*tmpbody = '\0';

			// save to file the first part of the line segment

			fprintf(newf,"%s<BR><PRE>\n%s</pre><BR>",line, dd->disclaimer_text);
			fprintf(newf,"<%s",(tmpbody+1));
			break;
		}
		else
		    {
			fprintf(newf,"%s",line);
		}

	} // While FFGET_fgets()

	return dd->html_inserted;
}




int AM_add_disclaimer_cleanup( FILE *mp, FILE *newf, char *mpacktmp, char *mpackname )
{

	// Close our files

	fclose(mp);
	fclose(newf);

	// rename the files

	rename(mpacktmp, mpackname);

	return 0;
}


/*------------------------------------------------------------------------
Procedure:     AM_add_disclaimer ID:1
Purpose:       Adds a disclaimer to [potentially] plain-text, mixed/multipart and HTML emails.
Input:         char *mpackname: Mailpack/email name we're modifying
char *disclaimer: either the text, or the filename of the text we want to insert
int istext: 0 = dislciamer is text, 1 = disclaimer is filename
Output:
Errors:
------------------------------------------------------------------------*/
int AM_add_disclaimer( char *mpackname, char *disclaimer, int istext )
{
	FILE *mp, *newf;
	FFGET_FILE f;
	char line[_BUFFER_MAX+1]="";
	char mpacktmp[_BUFFER_MAX+1]="";
	char mpackold[_BUFFER_MAX+1]="";
	struct AM_disclaimer_details dd;
	int result = 0;
	int last_boundary_written = -1;

	/* create our temp filename */
	snprintf(mpacktmp,_BUFFER_MAX, "%s.tmp",mpackname);
	snprintf(mpackold,_BUFFER_MAX, "%s.old",mpackname);

	/* Initialise our files */
	newf = fopen(mpacktmp,"w");
	mp = fopen(mpackname,"r");


	// Test file statuses... and hop out if we had troubles with either
	//
	if (!newf)
	{
		syslog(1,"AM_add_disclaimer: Cannot open %s, %s",mpacktmp,strerror(errno));
		return -1;
	}

	if (!mp)
	{
		syslog(1,"AM_add_disclaimer: Cannot open %s, %s",mpacktmp,strerror(errno));
		return -1;
	}


	// Set up the disclaimer text as required to be inserted into the file
	//

	if (istext == 0)
	{
		AM_disclaimer_load_text( disclaimer, &(dd.disclaimer_text) );
	}
	else
	    {
		dd.disclaimer_text = disclaimer;
	}


	// Initialise our variables

	dd.content_type = _none;
	dd.content_encoding = _none;
	dd.boundary_found = 0;
	memset(dd.boundary, 0, sizeof(dd.boundary));
	dd.ishtml = 0;
	dd.isfile = 0;
	dd.text_inserted = 0;
	dd.html_inserted = 0;


	if (__verbose) fprintf(stderr,"Adding disclaimer...\n");

	FFGET_setstream(&f, mp);

	// The process....
	//
	//	First, we read through the email until we at least come across the traditional
	//  true-blank seperator line between the headers and the body of the email.
	//  Once we have found that, we can then choose how we're going to detect our first
	//  disclaimer position based on what we discovered in the headers (ie, did we
	//  find a boundary specifier? are we using plain-text here, or does the email have
	//  HTML in it?
	//
	//  If we have just a plain text email, then we wait till we reach the end of the email
	//  and we just append the disclaimer
	//
	//  If we [most likely] dont have a plain email, then we need to wait until we find
	//  the boundary line, then, insert the disclaimer there.
	//
	//  Next, we need to check to see if the NEXT section is the HTML version of the first
	//  section (well, we dont /really/ compare them, but we check the headers to see if
	//  it has things like text/html and no filename
	//  If the section is HTML and non-filenamed, we once more wait until the boundary then
	//  we add in a HTML version of the disclaimer.
	//
	// Sounds easy ?
	//

	// Section 1 - read in the MAIN headers until a true-blank.
	//		This will give us some indication of what to expect in the email, be it a multipart
	//		a file, or a plain text email.
	//

	if (__verbose) fprintf(stderr,"Reading main headers\n");

	result = AM_add_section1( &dd, &f, newf );

	if ((AM_RETURN_SIGNED_EMAIL == result)||(-1 == result))
	{
		fclose(mp);
		fclose(newf);
		if (__verbose) fprintf(stderr,"Email is signed with a crypto, we do not alter these emails\n");
		return AM_RETURN_SIGNED_EMAIL;
	}


	// Section 2 - Look for the end of the first text-body
	//
	// Now, if we have no boundary on this email, and it's of
	// text/* content-type
	//

	// If we have no boundary, and we have either text or no content-type
	// then we can just append the disclaimer to the /end/ of the email

	if ((0 == dd.boundary_found)&&((dd.content_type == _text_plain)||(dd.content_type == _text_html)))
	{
		if (__verbose) fprintf(stderr,"Inserting disclaimer into a non-boundary email\n");

		if (dd.content_type == _text_html)
		{
			if (__verbose) fprintf(stderr,"Conditions right for HTML disclaimer to be added\n");

			result = AM_add_disclaimer_insert_html(  &dd, &f, newf );
			if ((__verbose)&&(0 == result))
			{
				fprintf(stderr,"WARNING: Could not insert HTML disclaimer into email\n");
			}

		}
		else
		    {
			while (FFGET_fgets(line, sizeof(line), &f))
			{
				fprintf(newf,"%s",line);
			}
			fprintf(newf,"%s",dd.disclaimer_text);
			dd.text_inserted = 1;
		}


	}


	// BOUNDARY LIMITED EMAILS -----------------
	//
	// These emails are somewhat /harder/ to insert attachments into
	// namely because of the profound number of convoulted ways in which
	// email clients can forward/reply/mangle emails.
	//
	// We use a 'WHILE' loop here just so that we can cheat a bit when
	// trying to /exit/ out of the processing prematurely when we run
	// out of input (as happens at the end of the input file.
	//
	// By using a 'WHILE', we can just use 'break', rather than resorting
	// to other ugly methods :)


	while (1 == dd.boundary_found)
	{
		if (FFGET_feof(&f)) break;

		// If we've found a boundary and a text content section...
		//

		if (__verbose) fprintf(stderr,"Inserting into a MIME email\n");


		// Step 1 - first read off any spurilous data prior to the first
		// 	MIME block, this typically includes things like
		//   "If you can read this, then your email client is not MIME
		//	  compliant"  (etc).

		if (__verbose) fprintf(stderr,"\tReading spurilous data before first MIME segment\n");

		while (FFGET_fgets(line, _BUFFER_MAX, &f))
		{
			fprintf(newf,"%s",line);
			if ( strstr(line,dd.boundary) ) break;
		}


		// IF for some reason we run into the end of the file, then we just quit out
		//	using our break.

		if (FFGET_feof(&f))
		{
			break;
		}


		// Okay, while we haven't yet managed to insert the disclaimer text
		//	or while we haven't inserted the HTML disclaimer text, we will
		//	keep going around in this loop

		while ((0 == dd.text_inserted)||(0 == dd.html_inserted))
		{

			// Step 2 - Now that we have the first block out of the way, we
			// can read to the bottom of the 2nd block, which is / should be
			// our text body


			// Read the headers and get the content-type information

			if (AM_add_section1( &dd, &f, newf ) < 0)
			{
				break;
				if (__verbose) fprintf(stderr,"\tRan out of input whilst processing headers\n");
			}

			// If this segment is a file, then rather than sifting through anything else
			//	just read until we get to the next boundary line and 'continue' on from
			//	the start of the WHILE loop again.

			if ( dd.isfile == 1 )
			{
				if (__verbose) fprintf(stdout,"\tJumping this MIME section as it is a FILE\n");

				while (FFGET_fgets(line, _BUFFER_MAX, &f))
				{

//					if (FFGET_feof( &f ))
//					{
//						fprintf(stdout,"debug: Lastline = --'%s'--", line );
//					}

					fprintf( newf, "%s", line );
					if ( strstr( line,dd.boundary ) ) break;
				}

				if (__verbose) fprintf(stdout,"\tFile writing DONE.\n");

				// IF for some reason we run into the end of the file, then we just quit out
				//	using our break.

				if (FFGET_feof(&f))
				{
					break;
				}

				continue;
			}


			// based on our findings from the header reading,
			//	If our MIME block is plain text...
			//
			// Also, take note that if the MIME block is for a file, we dont
			//	insert into it ( else our disclaimer will potentially become
			//	part of the file, really clever stuff!

			if ( (0 == dd.text_inserted) && (dd.content_type == _text_plain) && (dd.isfile == 0) )
			{
				if (__verbose) fprintf(stderr,"\tConditions right to insert disclaimer\n");

				// Read in the data (line by line) until we locate the
				// boundary line, then, we write out the disclaimer text
				// -before- we write the disclaimer line down.

				if (__verbose) fprintf(stderr,"\tReading first segment looking for boundary line\n");

				// Read and write data until we locate the boundary line.
				//	NOTE - we -deliberately- break before writing the boundary
				//	line because we want to insert the disclaimer /before/
				//	we write the boundary line.

				while (FFGET_fgets(line, _BUFFER_MAX, &f))
				{
					last_boundary_written = 0;
					if ( strstr( line,dd.boundary ) ) break;
					fprintf( newf, "%s", line );
					last_boundary_written = 1;
				}

				// If we got to the end of the file, then, well, ooops, I guess
				//	we'll have to break out.

				if ( FFGET_feof( &f ) )
				{
					break;
				}


				// Assuming that if the file is NOT encoded in base64, then we can attempt
				//	to insert the data. - This is a bit of a assumptious test, because we're
				//	going on the fact that our outer while-loop did a test to make sure that
				//	the current MIME section is NOT a file, therefore, we're saying
				//	"This segment is not a file and is not B64 encoded, so, we can now
				//	insert some disclaimer text into it"

				if (dd.content_encoding != _base64)
				{
					if (__verbose) fprintf(stderr,"\tInserting TEXT disclaimer (%s)\n", dd.disclaimer_text);

					fprintf( newf, "%s\n", dd.disclaimer_text );

					dd.text_inserted = 1; // Our disclaimer has been inserted, so set the flag

				} // if the content encoding is not B64

				if ( last_boundary_written == 0 )
				{
					fprintf( newf, "%s", line );
					last_boundary_written = 1;
				}

			} // If text not inserted and the segment is not a file


			// Add in the HTML disclaimer (if wanted)

			if ( ( __HTMLtoo ) && ( 0 == dd.html_inserted ) && ( dd.content_type == _text_html ) && ( dd.isfile == 0 ) )
			{
				if (__verbose) fprintf(stderr,"\tConditions right for HTML disclaimer to be added\n");

				result = AM_add_disclaimer_insert_html(  &dd, &f, newf );
				if ((__verbose)&&(0 == result))
				{
					fprintf(stderr,"WARNING: Could not insert HTML disclaimer into email\n");
				}


			} // If we also want to include a disclaimer in HTML body

		} // If conditions are right to insert our disclaimer into this email with boundaries




		// IF for some reason we run into the end of the file, then we just quit out
		//	using our break.

		if (FFGET_feof(&f))
		{

			if (__verbose) fprintf(stderr,"\tMain boundary outter-loop FEOF detector hit, breaking out\n");
			break;
		}

		// Clean up any remaining lines in the email

		while (FFGET_fgets(line, _BUFFER_MAX, &f)){
		    fprintf(newf,"%s",line);
//			fprintf(stdout,"DEBUG: %s",line );
		}

	} // If a boundary was found in the main headers


	// Clean up any remaining lines in the email


	if ( ! FFGET_feof( &f ) )
	{
		while (FFGET_fgets(line, _BUFFER_MAX, &f)) fprintf(newf,"%s",line);
	}


	// if we didn't get the disclaimer inserted, and this email has no boundary
	//	and there's no file-headers, then we can insert the plain text disclaimer
	//	at the end.


	if ( (!dd.text_inserted) && (!dd.html_inserted) && (!dd.boundary_found) && ( dd.isfile == 0 ) )
	{
		if (__verbose) fprintf(stderr,"\tNo disclaimer added, despite going through entire file...\n");

		fprintf(newf,"\n%s\n\n",dd.disclaimer_text);
	}

	if (__verbose) fprintf(stdout,"done.\n");


	AM_add_disclaimer_cleanup( mp, newf, mpacktmp, mpackname);


	// Free the malloc'd text if required
	if ((istext == 0)&&(dd.disclaimer_text))
	{
		free(dd.disclaimer_text);
	}

	return result;
}






/*------------------------------------------------------------------------
Procedure:     AM_filename_fix ID:1
Purpose:       Converts a MIME header
Input:         line: string with the filename
removed_prefix: filename prefix to use
removed_count: the number of attachments we've processed / ID
Output:
Errors:
------------------------------------------------------------------------*/
int AM_filename_fix( char *line, char *removed_prefix, int removed_count )
{
	char *pos;
	char lline[1024];

	/* process the current line, as it contains the filename, and replace the filename
	* with a -removed- one */
	pos = strrchr(line,'='); /* locate where the filename starts */
	pos++; /* move one char along */
	*pos = '\0'; /* terminate the string */

	sprintf(lline,"%s\"%s%d\"\n",line,removed_prefix,removed_count); /* create a new string, in lline (scratch pad)*/
	strcpy(line,lline);

	return 0;
}






/*------------------------------------------------------------------------
Procedure:     AM_nullify_attachment ID:1
Purpose:       Removes any attachment within the mailpack which matches
(minimally) with the attachmentname parameter
Input:         mpackname - The mailpack name
attachmentname - The name of attachments which is to be removed
Output:        none
Errors:        If attachment wasn't found, or could not be removed, a non-zero return value.
------------------------------------------------------------------------*/
int AM_nullify_attachment( char *mpackname, char *attachmentname, char *removed_prefix )
{
	FILE *mp, *newf;
	char line[1024]="";
	char safe[1024]="";
	char lline[1024]="";
	char mpacktmp[1024]="";
	char mpackold[1024]="";
	char boundary[1024]="";

	int content_type = _none;
	int content_encoding = _none;

	int result = 0;
	int boundary_found = 0;
	int filename_found = 0;
	int removed_count = 0;
	char *pos, *newpos;

	if (__verbose) fprintf(stdout,"Removing attachments \"%s\"...",attachmentname);

	/* create our temp filename */
	sprintf(mpacktmp,"%s.tmp",mpackname);
	sprintf(mpackold,"%s.old",mpackname);

	/* Initialise our files */
	newf = fopen(mpacktmp,"w");

	if (!newf)
	{
		if (_AM_syslogging > 0) syslog(1,"AM_nullifyattachment: Error: Cannot open temporary file for writing (%s) : %s",mpacktmp, strerror(errno));
		if (_AM_stderrlogging > 0) fprintf(stderr,"AM_nullifyattachment: Error: Cannot open temporary file for writing");
		return -1;
	}

	mp = fopen(mpackname,"r");
	if (!mp)
	{
		if (_AM_syslogging > 0) syslog(1,"AM_nullifyattachment: Error: Cannot open temporary file for reading (%s) : %s",mpackname, strerror(errno));
		if (_AM_stderrlogging > 0) fprintf(stderr,"AM_nullifyattachment: Error: Cannot open temporary file for reading");
		return -1;
	}

	/* Initialise our variables */
	content_type = _none;
	content_encoding = _none;

	AM_strlower(attachmentname);

	/* Read every line from the original mailpack */
	while ( fgets(line,1024,mp) )
	{

		/* copy and convert into lower-case the input line */
		strcpy(lline,line);
		strcpy(safe,line);
		AM_strlower(lline);

		/* check and see if we've found the filename... */

		if ( (pos=strstr(lline, "name="))!=NULL)
		{
			if (  (__nullify_all) || (pos=strstr(lline,attachmentname))  )
			{
				filename_found = 1;
				if (__verbose) fprintf(stdout,"\nlocated attachment...");
			}
		}

		/* check and see if this is a content-type line */
		if ((pos=strstr(lline, "content-type:")))
		{
			pos += strlen("content-type:");	/* move our pointer along */

			if ((pos=strstr(pos,"text/")))	/* do we have a text type */
			{
				pos += strlen("text/");	/* move pointer along */

				if (strncmp(pos, "plain",5)==0) content_type = _text_plain;
				else
				    if (strncmp(pos, "html",4)==0) content_type = _text_html;
				else
				    if (strncmp(pos, "print",5)==0) content_type = _text_print;
				else content_type = _text_unknown;

			}
			else {
				content_type = _unknown;
			}
		}

		/* Our transfer encoding system indicates if we have a attachment type which is
		* binary-encoded stuff or plain text */
		if ((pos=strstr(lline, "content-transfer-encoding:")))
		{
			pos += strlen("content-transfer-encoding:");
			if (strstr(pos,"base64")) content_encoding = _base64;
			else if (strstr(pos,"quote")) content_encoding = _print;
			else if (strstr(pos,"7bit")) content_encoding = _7bit;
		}


		if (boundary_found < 1)
		{ /* if we have not found our boundary yet */
			pos = strstr(lline, "boundary=");	/* check for our boundary */
			if (pos) { /* if we found one! */

				newpos = (pos -lline) +line +strlen("boundary=");	/* move the pos pointer to just after the = */
				pos = newpos;
				if (*pos == '\"') pos++;	/* if our 1st char is a quote, move one more */
				newpos = strtok(pos,"\n\r\"");	/* tokenise out the boundary */
				strcpy(boundary, newpos);	/* copy it to our static string array */
				boundary_found++;	/* increment the boundary_found flag */
				if (__verbose) fprintf(stdout,"found boundary...");
			} /* if POS */
		} /* if boundary_found < 1 */

		else
		    { /* if boundary_found >= 1 */

			/* if we've found the filename, then we must process until we find the next boundary */
			if (filename_found == 1)
			{
				AM_filename_fix(line, removed_prefix, removed_count);
				fprintf(newf,"%s",line); /* output the new filename */
				if (__verbose) fprintf(stdout,"renamed attachment...");

				/* first, locate our \n\r blank line */

				while ( (fgets(line, 1024,mp)) && (strlen(line) >3) ){
					strcpy(lline,line);
					AM_strlower(lline);

					if (strstr(lline,"name=")){
						AM_filename_fix(line, removed_prefix, removed_count);
					}

					fprintf(newf,"%s",line);
				}


				/* Now, until the boundary is found again, absorb all lines */
				while ( (fgets(line, 1024,mp)) && (!strstr(line, boundary)) );

				/* leave an empty file */
				fprintf(newf,"\n\n");

				/* and, given that we will exit on the boundary line, we can just edit from this loop as we write it later */
				strcpy(safe, line);

				/* we are -done- */
				filename_found = 0;
				removed_count++;
				if (__verbose) fprintf(stdout,"nullified.");
			} /* if filename_found == 1 */

		} /* if-ELSE */

		fprintf(newf,"%s",safe); /* save the original line to the output file */
		filename_found = 0;

	} /* WHILE more lines to read in from original */

	if (__verbose) fprintf(stdout,"\ndone.\n");
	/* Close our files */
	fclose(mp);
	fclose(newf);

	/* Rename files */
	rename(mpackname, mpackold);
	rename(mpacktmp, mpackname);

	return result;
}


/*------------------------------------------------------------------------
Procedure:     AM_insert_Xheader ID:1
Purpose:       Adds an X-header to the first set of headers in an email mailpack.
Input:         char *fname: mailpack name
char *xheader: header string
Output:
Errors:
------------------------------------------------------------------------*/
int AM_insert_Xheader( char *fname, char *xheader)
{

	/* Tempfile tmpfile() fix contributed by David DeMaagd - 29/01/2001 */

	char line[_BUFFER_MAX +1];
	char tpn[_BUFFER_MAX +1];
	int header_written = 0;
	int result = 0;
	struct stat st;
	FFGET_FILE f;
	FILE *fi;
	FILE *fo;

	// Create a temporary file name, but, so that we dont
	// overwrite an existing file, we must check the name using
	// stat() to see if we get a non-zero response.

	snprintf(tpn, _BUFFER_MAX, "%s",fname);

	do {
		if (strlen(tpn) < (sizeof(tpn) -2))
		{
			strcat(tpn,".");
		}
		else
		    {
			fprintf(stderr,"ERROR:%s:%d: Temporary file name string buffer out of space!\n",__FILE__,__LINE__);
			return 1;
		}
		//		fprintf(stderr,"DEBUG:%s:%d: Testing filename %s\n",__FILE__,__LINE__,tpn);
												}
	while (0 == stat(tpn,&st));


	// Attempt to open up the temporary file in write mode.
	// If the open fails, then this whole operation fails.
	// Ensure there's lots of good logging here so that when something
	// does go wrong, at least people wont be left in the dark as to
	// what went on.
	//
	// Same applies for opening the source file
	//

	fo = fopen(tpn,"w");
	if (!fo)
	{
		if (_AM_syslogging > 0) syslog(1,"Error: Cannot open temporary file for writing (%s) : %s",tpn, strerror(errno));
		if (_AM_stderrlogging > 0) fprintf(stderr,"Error: Cannot open temporary file for writing");
		return 1;
	}


	fi = fopen(fname,"r");
	if (!fi)
	{
		if (_AM_syslogging > 0) syslog(1,"Error: Cannot open Mailpack file for reading (%s) : %s",fname, strerror(errno));
		if (_AM_stderrlogging > 0) fprintf(stderr,"Error: Cannot open Mailpack file for reading, %s",fname);
		return 2;
	}

	//setup our FFGET stream

	FFGET_setstream(&f, fi);


	// Load and go through every line in the email file, testing it for "trueblank"
	// status.  Once a trueblank is found, we write our header, as this marks the
	// end of the headers (the trueblank).  Beyond that, we keep on reading and
	// writing because we obviously want to keep the entire email content.

	while (FFGET_fgets(line, _BUFFER_MAX ,&f))
	{

		// If we've found that the line we just read is a /true/ blank (as apposed to a carefully
		// crafted line length to catch out our fgets(), /and/ if we've not written in a header
		// then we will insert the header.

		if ((0 != f.trueblank)&&(0 == header_written))
		{
			fprintf(fo,"%s\n",xheader);
			header_written = 1;
		}

		// In all cases, we must write out the line we read in, even if it was the true-blank line
		// Otherwise we'll end up losing a the header seperation, and all of a sudden your emails
		// do not make any more sence to email programs.

		fprintf(fo,"%s",line);
	}

	// Close our input files


	fclose(fo);
	fclose(fi);

	// We now can rename the temporary file to the new file name.
	// NOTE - previously we removed then renamed, however, on checking
	// with the manpages, 'man 2 rename' it seems it's not required, so
	// rather than waste CPU cycles, we'll just go by the book, and use
	// only rename.  There are a couple of situations where rename can/will
	// fail, obviously such as if the original file is marked read-only, or
	// we do not have write permissions to it.


	if (rename(tpn,fname) == -1)
	{
		result = 1;
		if (_AM_stderrlogging > 0) fprintf(stderr,"Error: xInsertHeader - RENAME, %s",strerror(errno));
	}

	return result;
}





//-----------------------------------END----


